/*
 * Copyright (c) 2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include <sys/time.h>

#include <math.h>

#include "mrsim_c_types.h"
#include "codels.h"


/* --- Task main -------------------------------------------------------- */


/** Codel mrsim_main_init of task main.
 *
 * Triggered by mrsim_start.
 * Yields to mrsim_main.
 */
genom_event
mrsim_main_init(mrsim_ids *ids, const mrsim_gps *gps,
                const mrsim_state *state, genom_context self)
{
  or_pose_estimator_state *sdata, *gdata;
  struct timeval tv;
  int i;

  static const double kf = 6.5e-4;
  static const double c = 0.0154;
  static const double d = 0.23;

  gettimeofday(&tv, NULL);

  ids->simparam = (mrsim_ids_simparam_s) {
    .motor = {
      .J = 8e-5,
      .b = 5e-4,
      .K = 0.01,
      .R = 0.2,
      .L = 2e-4,
      .V = 16.
    },

    .body = {
      .G = {
        0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0.,
        kf, kf, kf, kf, 0., 0., 0., 0.,

        0.,     d*kf,   0., -d*kf, 0., 0., 0., 0.,
        -d*kf,    0., d*kf,    0., 0., 0., 0., 0.,
         c*kf, -c*kf, c*kf, -c*kf, 0., 0., 0., 0.
      },

      .J = {
        0.015,    0.,    0.,
        0.,    0.015,    0.,
        0.,       0., 0.015
      },

      .m = 1.0
    },

    .state = {
      .pstddev = 1e-3, .qstddev = 5e-3,
      .vstddev = 1e-3, .wstddev = 5e-3,
      .astddev = 2e-2,

      .imup = 0., .batp = 0., .motp = 0., .gpsp = .01
    }
  };

  ids->simstate = (mrsim_ids_simstate_s) {
    .ts = { .sec = tv.tv_sec, .nsec = 1000 * tv.tv_usec },
    .imuts = { .sec = tv.tv_sec, .nsec = 1000 * tv.tv_usec },
    .batts = { .sec = tv.tv_sec, .nsec = 1000 * tv.tv_usec },
    .motts = { .sec = tv.tv_sec, .nsec = 1000 * tv.tv_usec },
  };
  for(i = 0; i < or_rotorcraft_max_rotors; i++) {
    ids->simstate.motor[i] = (mrsim_ids_simstate_s_motor_s) {
      .emerg = false, .spinning = false, .starting = false,
      .disabled = i < 4 ? false : true,
      .pwm = 0., .w = 0., .i = 0.
    };
  }

  ids->pty = (mrsim_ids_pty_s){
    .fd = -1,
    .device = "/tmp/mikrokopter"
  };

  sdata = state->data(self);
  double pvar = ids->simparam.state.pstddev * ids->simparam.state.pstddev;
  double qvar = ids->simparam.state.qstddev * ids->simparam.state.qstddev;
  double vvar = ids->simparam.state.vstddev * ids->simparam.state.vstddev;
  double wvar = ids->simparam.state.wstddev * ids->simparam.state.wstddev;
  double avar = ids->simparam.state.astddev * ids->simparam.state.astddev;

  *sdata = (or_pose_estimator_state){
    .ts = { 0, 0 },
    .intrinsic = false,

    .pos = {
      ._present = true,
      ._value = {
        .x = 0., .y = 0., .z = 0.,
        .qw = 1., .qx = 0., .qy = 0., .qz = 0.
      }
    },
    .pos_cov = {
      ._present = true,
      ._value = {{
          pvar,
          0., pvar,
          0.,   0., pvar,
          0.,   0.,   0., 0.,
          0.,   0.,   0., 0., qvar,
          0.,   0.,   0., 0.,   0., qvar,
          0.,   0.,   0., 0.,   0.,   0., qvar
        }}
    },

    .vel = {
      ._present = true,
      ._value = {
        .vx = 0., .vy = 0., .vz = 0.,
        .wx = 0., .wy = 0., .wz = 0.
      }
    },
    .vel_cov = {
      ._present = true,
      ._value = {{
          vvar,
          0., vvar,
          0.,   0., vvar,
          0.,   0.,   0., wvar,
          0.,   0.,   0.,   0., wvar,
          0.,   0.,   0.,   0.,   0., wvar
        }}
    },

    .acc = {
      ._present = true,
      ._value = {
        .ax = 0., .ay = 0., .az = 0.
      }
    },
    .acc_cov = {
      ._present = true,
      ._value = {{
          avar,
          0., avar,
          0.,   0., avar
        }}
    },
  };

  gdata = gps->data(self);
  *gdata = (or_pose_estimator_state){
    .ts = sdata->ts,
    .intrinsic = false,

    .pos = sdata->pos,
    .pos_cov = sdata->pos_cov,

    .vel = { ._present = false },
    .vel_cov = { ._present = false },

    .acc = { ._present = false },
    .acc_cov = { ._present = false }
  };

  return mrsim_main;
}


/** Codel mrsim_main_perm of task main.
 *
 * Triggered by mrsim_main.
 * Yields to mrsim_pause_main.
 */
genom_event
mrsim_main_perm(const mrsim_ids_simparam_s *simparam,
                const mrsim_ids_pty_s *pty,
                mrsim_ids_simstate_s *simstate, const mrsim_gps *gps,
                const mrsim_state *state, genom_context self)
{
  or_pose_estimator_state *sdata = state->data(self);

  struct timeval tv;
  double dt;
  int i;

  /* update time */
  gettimeofday(&tv, NULL);
  dt =
    tv.tv_sec - simstate->ts.sec + 1e-6 * tv.tv_usec - 1e-9 * simstate->ts.nsec;
  simstate->ts.sec = tv.tv_sec;
  simstate->ts.nsec = 1000 * tv.tv_usec;

  /* update motors */
  for(i = 0; i < or_rotorcraft_max_rotors; i++) {
    if (simstate->motor[i].disabled) continue;

    /* watchdog */
    if (simstate->ts.sec - simstate->motor[i].wdts.sec
        + 1e-9 * (simstate->ts.nsec - simstate->motor[i].wdts.nsec) > 0.5)
      simstate->motor[i].wd = 1e6/65535;

    /* emulate the variable period of the BL control algorithm */
    int it, niter;
    if (fabs(simstate->motor[i].w) < 25)
      niter = 1 + dt*1000;
    else
      niter = 1 + dt * 42*fabs(simstate->motor[i].w);

    for(it = 0; it < niter; it++) {
      mrsim_motor_control(&simparam->motor, &simstate->motor[i]);
      mrsim_motor_model(&simparam->motor, &simstate->motor[i], dt/niter);
    }
  }


  /* update state */
  mrsim_wrench(&simparam->body, simstate);
  mrsim_dynamics(&simparam->body, &simstate->body, sdata, dt);

  {
    double qvar = simparam->state.qstddev * simparam->state.qstddev;
    double qw = sdata->pos._value.qw;
    double qx = sdata->pos._value.qx;
    double qy = sdata->pos._value.qy;
    double qz = sdata->pos._value.qz;

    sdata->pos_cov._value.cov[9]  = qvar * (1 - qw*qw);
    sdata->pos_cov._value.cov[13] = qvar * -qw*qx;
    sdata->pos_cov._value.cov[14] = qvar * (1 - qx*qx);
    sdata->pos_cov._value.cov[18] = qvar * qw*qy;
    sdata->pos_cov._value.cov[19] = qvar * -qx*qy;
    sdata->pos_cov._value.cov[20] = qvar * (1 - qy*qy);
    sdata->pos_cov._value.cov[24] = qvar * -qw*qz;
    sdata->pos_cov._value.cov[25] = qvar * -qx*qz;
    sdata->pos_cov._value.cov[26] = qvar * -qy*qz;
    sdata->pos_cov._value.cov[27] = qvar * (1 - qz*qz);
  }
  gps->data(self)->pos = sdata->pos;
  gps->data(self)->pos_cov = sdata->pos_cov;

  mrsim_addnoise(&simparam->state, gps->data(self));


  /* publish or perish */
  sdata->ts = simstate->ts;
  state->write(self);

  if (simstate->ts.sec - gps->data(self)->ts.sec
      + 1e-9 * (simstate->ts.nsec - gps->data(self)->ts.nsec) >
      0.9 * simparam->state.gpsp) {
    gps->data(self)->ts = simstate->ts;
    gps->write(self);
  }

  if (simparam->state.imup > 0 &&
      simstate->ts.sec - simstate->imuts.sec
      + 1e-9 * (simstate->ts.nsec - simstate->imuts.nsec) >
      0.9 * simparam->state.imup) {
    simstate->imuts = simstate->ts;
    mrsim_send_msg(
      pty->fd, "I%1%2%2%2%2%2%2",
      mrsim_seqid(tv, simparam->state.imup),
      (int16_t)(1000.*(simstate->body.a[0] + gnoise(simparam->state.astddev))),
      (int16_t)(1000.*(simstate->body.a[1] + gnoise(simparam->state.astddev))),
      (int16_t)(1000.*(simstate->body.a[2] + gnoise(simparam->state.astddev))),
      (int16_t)(1000.*(simstate->body.w[0] + gnoise(simparam->state.wstddev))),
      (int16_t)(1000.*(simstate->body.w[1] + gnoise(simparam->state.wstddev))),
      (int16_t)(1000.*(simstate->body.w[2] + gnoise(simparam->state.wstddev))));
  }

  if (simparam->state.batp > 0 &&
      simstate->ts.sec - simstate->batts.sec
      + 1e-9 * (simstate->ts.nsec - simstate->batts.nsec) >
      0.9 * simparam->state.batp) {
    simstate->batts = simstate->ts;
    mrsim_send_msg(
      pty->fd, "B%1%2",
      mrsim_seqid(tv, simparam->state.batp),
      (uint16_t)(1000.*simparam->motor.V));
  }

  if (simparam->state.motp > 0 &&
      simstate->ts.sec - simstate->motts.sec
      + 1e-9 * (simstate->ts.nsec - simstate->motts.nsec) >
      0.9 * simparam->state.motp) {
    simstate->motts = simstate->ts;

    for(i = 0; i < or_rotorcraft_max_rotors; i++) {
      if (simstate->motor[i].disabled) continue;

      double w = mrsim_motor_measure(&simstate->motor[i]);
      int16_t p;
      uint8_t s;

      s = i + 1;
      if (simstate->motor[i].emerg) s |= 0x80;
      if (simstate->motor[i].spinning) s |= 0x20;
      if (simstate->motor[i].starting) s |= 0x10;

      p = (fabs(w) > 1e6/65535) ? 1e6/2/w : copysign(32767, w);

      mrsim_send_msg(
        pty->fd, "M%1%1%2%2%2",
        mrsim_seqid(tv, simparam->state.motp), s, p,
        (int16_t)(1024.*copysign(simstate->motor[i].pwm, w)),
        (uint16_t)(1000.*simstate->motor[i].i));
    }
  }

  return mrsim_pause_main;
}
