/*
 * Copyright (c) 2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include <err.h>

#include "mrsim_c_types.h"
#include "codels.h"


/* --- Task read -------------------------------------------------------- */

static struct mrsim_pty_s chan;

/** Codel mrsim_read_init of task read.
 *
 * Triggered by mrsim_start.
 * Yields to mrsim_main.
 */
genom_event
mrsim_read_init(mrsim_ids_pty_s *pty, genom_context self)
{
  if (pty->fd >=0) mrsim_close_pty(pty->fd, pty->device);

  pty->fd = mrsim_open_pty(pty->device);
  if (pty->fd < 0) return mrsim_e_sys_error("pty", self);

  chan.r = chan.w = 0;
  chan.start = chan.escape = false;
  warnx("created %s", pty->device);
  return mrsim_main;
}


/** Codel mrsim_read_perm of task read.
 *
 * Triggered by mrsim_main.
 * Yields to mrsim_recv, mrsim_main, mrsim_start.
 */
genom_event
mrsim_read_perm(const mrsim_ids_pty_s *pty, genom_context self)
{
  int s;

  s = mrsim_wait_msg(pty->fd);
  if (s < 0) {
    warn("pty");
    return mrsim_start;
  }

  return s == 0 ? mrsim_main : mrsim_recv;
}


/** Codel mrsim_read_recv of task read.
 *
 * Triggered by mrsim_recv.
 * Yields to mrsim_main.
 */
genom_event
mrsim_read_recv(const mrsim_ids_pty_s *pty,
                mrsim_ids_simparam_s_state_s *state,
                mrsim_ids_simstate_s_motor_s motor[8],
                genom_context self)
{
  if (mrsim_recv_msg(pty->fd, &chan) <= 0) return mrsim_main;

  switch(chan.msg[0]) {
    case 'i': /* IMU data */
      if (chan.len == 5) {
        uint32_t p =
          ((uint32_t)chan.msg[1] << 24) |
          ((uint32_t)chan.msg[2] << 16) |
          ((uint32_t)chan.msg[3] << 8) |
          ((uint32_t)chan.msg[4] << 0);

        state->imup = p/1e6;
      }
      break;

    case 'b': /* battery level */
      if (chan.len == 5) {
        uint32_t p =
          ((uint32_t)chan.msg[1] << 24) |
          ((uint32_t)chan.msg[2] << 16) |
          ((uint32_t)chan.msg[3] << 8) |
          ((uint32_t)chan.msg[4] << 0);

        state->batp = p/1e6;
      }
      break;

    case 'm': /* motor data */
      if (chan.len == 5) {
        uint32_t p =
          ((uint32_t)chan.msg[1] << 24) |
          ((uint32_t)chan.msg[2] << 16) |
          ((uint32_t)chan.msg[3] << 8) |
          ((uint32_t)chan.msg[4] << 0);

        state->motp = p/1e6;
      }
      break;

    case 'g': /* start motors */ {
      int i;

      for(i = 0; i < or_rotorcraft_max_rotors; i++) {
        if (chan.len > 1 && chan.msg[1] != i + 1) continue;
        if (motor[i].disabled) continue;
        if (motor[i].spinning) continue;

        motor[i].starting = true;
        motor[i].spinning = false;
        motor[i].servo = false;
      }
      break;
    }

    case 'x': /* stop motors */ {
      int i;

      for(i = 0; i < or_rotorcraft_max_rotors; i++) {
        if (chan.len > 1 && chan.msg[1] != i + 1) continue;
        if (motor[i].disabled) continue;

        motor[i].starting = false;
        motor[i].spinning = false;
        motor[i].servo = false;
      }
      break;
    }

    case 'w': /* velocity array */ {
      int i;
      int32_t p;

      for(i = 0; i < or_rotorcraft_max_rotors; i++) {
        if (chan.len < 3 + 2 * i) break;

        p =
          ((int32_t)chan.msg[1 + 2 * i] << 9) |
          ((uint32_t)chan.msg[2 + 2 * i] << 1);
        mrsim_motor_period(motor + i, 1e6/p);
      }
      break;
    }

    case '?': /* id */
      if (chan.len == 1) {
        mrsim_send_msg(pty->fd, "?mkfl1.7");
      }
      break;
  }

  return mrsim_recv;
}


/** Codel mrsim_read_stop of task read.
 *
 * Triggered by mrsim_stop.
 * Yields to mrsim_ether.
 */
genom_event
mrsim_read_stop(const mrsim_ids_pty_s *pty, genom_context self)
{
  if (pty->fd >=0) mrsim_close_pty(pty->fd, pty->device);
  return mrsim_ether;
}
