/*
 * Copyright (c) 2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "acmrsim.h"

#include <cmath>

#include <Eigen/Geometry>

#include "mrsim_c_types.h"
#include "codels.h"

/* --- mrsim_motor_model --------------------------------------------------- */

static void
mrsim_motor_eq(const struct mrsim_ids_simparam_s_motor_s *motor,
               const double pwm, const double w, const double i,
               double *dw, double *di)
{
  *dw =
    - motor->b * w
    + motor->K * i / (2*M_PI);
  *di =
    - 2*M_PI * motor->K * w
    - motor->R * i
    + pwm * motor->V;
}


void
mrsim_motor_model(const struct mrsim_ids_simparam_s_motor_s *motor,
                  struct mrsim_ids_simstate_s_motor_s *state,
                  const double dt)
{
  struct { double dw, di; } k[4];

  /* RK4 */
  mrsim_motor_eq(motor, state->pwm,
                 state->w, state->i, &k[0].dw, &k[0].di);
  mrsim_motor_eq(motor, state->pwm,
                 state->w + dt/motor->J/2 * k[0].dw,
                 state->i + dt/motor->L/2 * k[0].di, &k[1].dw, &k[1].di);
  mrsim_motor_eq(motor, state->pwm,
                 state->w + dt/motor->J/2 * k[1].dw,
                 state->i + dt/motor->L/2 * k[1].di, &k[2].dw, &k[2].di);
  mrsim_motor_eq(motor, state->pwm,
                 state->w + dt/motor->J * k[2].dw,
                 state->i + dt/motor->L * k[2].di, &k[3].dw, &k[3].di);

  state->w += dt/motor->J/6 * (k[0].dw + 2*k[1].dw + 2*k[2].dw + k[3].dw);
  state->i += dt/motor->L/6 * (k[0].di + 2*k[1].di + 2*k[2].di + k[3].di);
  if (state->i > 40) state->i = 40;
}


/* --- mrsim_wrench -------------------------------------------------------- */

void
mrsim_wrench(const struct mrsim_ids_simparam_s_body_s *body,
             struct mrsim_ids_simstate_s *state)
{
  int i, j;

  for(i = 0; i < 3; i++) {
    state->body.thrust[i] = 0.;
    for(j = 0; j < or_rotorcraft_max_rotors; j++)
      state->body.thrust[i] +=
        body->G[i*or_rotorcraft_max_rotors + j] *
        state->motor[j].w * state->motor[j].w;

    state->body.torque[i] = 0.;
    for(j = 0; j < or_rotorcraft_max_rotors; j++)
      state->body.torque[i] +=
        body->G[(i+3)*or_rotorcraft_max_rotors + j] *
        state->motor[j].w * state->motor[j].w;
  }
}


/* --- mrsim_dynamics ------------------------------------------------------ */

void
mrsim_dynamics(const struct mrsim_ids_simparam_s_body_s *bodyp,
               struct mrsim_ids_simstate_s_body_s *body,
               or_pose_estimator_state *state, double dt)
{
  const Eigen::Matrix3d J(bodyp->J);
  const Eigen::Vector3d thrust(body->thrust);
  const Eigen::Vector3d torque(body->torque);
  static const Eigen::Vector3d g(0, 0, -9.81);

  /* current state */
  Eigen::Vector3d p(
    state->pos._value.x, state->pos._value.y, state->pos._value.z);
  Eigen::Quaternion<double> q(
    state->pos._value.qw,
    state->pos._value.qx, state->pos._value.qy, state->pos._value.qz);
  Eigen::Vector3d v(
    state->vel._value.vx, state->vel._value.vy, state->vel._value.vz);
  Eigen::Vector3d w(
    state->vel._value.wx, state->vel._value.wy, state->vel._value.wz);
  Eigen::Vector3d a =
    g + q._transformVector(thrust) / bodyp->m;

  /* current state in body frame */
  Eigen::Map<Eigen::Vector3d> wb(body->w);
  Eigen::Vector3d dwb;
  Eigen::Map<Eigen::Vector3d> ab(body->a);

  wb = q.conjugate()._transformVector(w);
  dwb = J.inverse()*((J*wb).cross(wb) + torque);
  ab = q.conjugate()._transformVector(a - g);

  /* update linear velocity */
  v += a * dt;

  /* update position */
  p += v * dt + a * dt*dt/2;

  /* update angular velocity */
  wb += dwb * dt;
  w = q._transformVector(wb);

  /* update attitude */
  Eigen::Quaternion<double> dq;
  double a2 = dt * dt * w.squaredNorm();
  if (a2 < 1e-1) {
    dq.w() = 1 - a2/8 /*std::cos(a/2)*/;
    dq.vec() = (0.5 - a2/48 /*std::sin(a/2)/a*/) * dt * w;
  } else {
    double a = std::sqrt(a2);
    dq.w() = std::cos(a/2);
    dq.vec() = std::sin(a/2)/a * dt * w;
  }
  q = dq * q;

  /* ground reaction */
  if (p(2) < 0.) {
    p(2) = 0.;
    q.w() = sqrt(1 - q.vec()(2)*q.vec()(2));
    q.vec() << 0., 0., q.vec()(2);
    v << 0., 0., 0.;
    w << 0., 0., 0.;
    a << 0., 0., 0.;
    ab = -g;
    wb << 0., 0., 0.;
  }

  /* update state */
  state->pos._value.x = p(0);
  state->pos._value.y = p(1);
  state->pos._value.z = p(2);
  state->pos._value.qw = q.w();
  state->pos._value.qx = q.vec()(0);
  state->pos._value.qy = q.vec()(1);
  state->pos._value.qz = q.vec()(2);

  state->vel._value.vx = v(0);
  state->vel._value.vy = v(1);
  state->vel._value.vz = v(2);
  state->vel._value.wx = w(0);
  state->vel._value.wy = w(1);
  state->vel._value.wz = w(2);

  state->acc._value.ax = a(0);
  state->acc._value.ay = a(1);
  state->acc._value.az = a(2);
}


/* --- mrsim_gnoise -------------------------------------------------------- */

double
gnoise(double stddev)
{
  double u1, u2;
  double s;

  do {
    u1 = 2 * drand48() - 1; /* [-1; 1] */
    u2 = 2 * drand48() - 1;

    s = u1*u1 + u2*u2;
  } while(s >= 1.);

  return sqrt(-2. * log(s)/s) * stddev * u1;
}

/* --- mrsim_addnoise ------------------------------------------------------ */

void
mrsim_addnoise(const struct mrsim_ids_simparam_s_state_s *param,
               or_pose_estimator_state *state)
{
  if (state->pos._present) {
    if (param->pstddev > 0.) {
      state->pos._value.x += gnoise(param->pstddev);
      state->pos._value.y += gnoise(param->pstddev);
      state->pos._value.z += gnoise(param->pstddev);
    }

    if (param->qstddev > 0.) {
      double nw;
      double nx = gnoise(param->qstddev);
      double ny = gnoise(param->qstddev);
      double nz = gnoise(param->qstddev);
      double n = sqrt(nx*nx + ny*ny + nz*nz);
      if (n > 0.) {
        double qw = state->pos._value.qw;
        double qx = state->pos._value.qx;
        double qy = state->pos._value.qy;
        double qz = state->pos._value.qz;

        nw = cos(n/2);
        nx *= sin(n/2)/n;
        ny *= sin(n/2)/n;
        nz *= sin(n/2)/n;

        state->pos._value.qw = qw*nw - qx*nx - qy*ny - qz*nz;
        state->pos._value.qx = qw*nx + qx*nw + qy*nz + qz*ny;
        state->pos._value.qy = qw*ny - qx*nz + qy*nw + qz*nx;
        state->pos._value.qz = qw*nz + qx*ny - qy*nx + qz*nw;
      }
    }
  }

  if (state->vel._present && param->wstddev > 0.) {
    state->vel._value.wx += gnoise(param->wstddev);
    state->vel._value.wy += gnoise(param->wstddev);
    state->vel._value.wz += gnoise(param->wstddev);
  }

  if (state->acc._present && param->astddev > 0.) {
    state->acc._value.ax += gnoise(param->astddev);
    state->acc._value.ay += gnoise(param->astddev);
    state->acc._value.az += gnoise(param->astddev);
  }
}
